# This file was auto-generated by Fern from our API Definition.

from .resources import (
    AccessDeniedError,
    CreateEventRequest,
    CreateLog,
    CreateScoreRequest,
    CreateSpanRequest,
    CreateTraceRequest,
    Error,
    Event,
    LlmUsage,
    Log,
    MapValue,
    MethodNotAllowedError,
    ObservationLevelEvent,
    ObservationLevelGeneration,
    ObservationLevelSpan,
    Score,
    Span,
    Trace,
    TraceIdType,
    TraceIdTypeEvent,
    TraceIdTypeGenerations,
    TraceIdTypeSpan,
    UnauthorizedError,
    UpdateGenerationRequest,
    UpdateSpanRequest,
    commons,
    event,
    generations,
    score,
    span,
    trace,
)

__all__ = [
    "AccessDeniedError",
    "CreateEventRequest",
    "CreateLog",
    "CreateScoreRequest",
    "CreateSpanRequest",
    "CreateTraceRequest",
    "Error",
    "Event",
    "LlmUsage",
    "Log",
    "MapValue",
    "MethodNotAllowedError",
    "ObservationLevelEvent",
    "ObservationLevelGeneration",
    "ObservationLevelSpan",
    "Score",
    "Span",
    "Trace",
    "TraceIdType",
    "TraceIdTypeEvent",
    "TraceIdTypeGenerations",
    "TraceIdTypeSpan",
    "UnauthorizedError",
    "UpdateGenerationRequest",
    "UpdateSpanRequest",
    "commons",
    "event",
    "generations",
    "score",
    "span",
    "trace",
]



# This file was auto-generated by Fern from our API Definition.

from .types import (
    CreateLog,
    LlmUsage,
    Log,
    MapValue,
    ObservationLevelGeneration,
    TraceIdTypeGenerations,
    UpdateGenerationRequest,
)

__all__ = [
    "CreateLog",
    "LlmUsage",
    "Log",
    "MapValue",
    "ObservationLevelGeneration",
    "TraceIdTypeGenerations",
    "UpdateGenerationRequest",
]



# This file was auto-generated by Fern from our API Definition.

from .types import CreateEventRequest, Event, ObservationLevelEvent, TraceIdTypeEvent

__all__ = ["CreateEventRequest", "Event", "ObservationLevelEvent", "TraceIdTypeEvent"]



# This file was auto-generated by Fern from our API Definition.

from . import commons, event, generations, score, span, trace
from .commons import AccessDeniedError, Error, MethodNotAllowedError, UnauthorizedError
from .event import CreateEventRequest, Event, ObservationLevelEvent, TraceIdTypeEvent
from .generations import (
    CreateLog,
    LlmUsage,
    Log,
    MapValue,
    ObservationLevelGeneration,
    TraceIdTypeGenerations,
    UpdateGenerationRequest,
)
from .score import CreateScoreRequest, Score, TraceIdType
from .span import CreateSpanRequest, ObservationLevelSpan, Span, TraceIdTypeSpan, UpdateSpanRequest
from .trace import CreateTraceRequest, Trace

__all__ = [
    "AccessDeniedError",
    "CreateEventRequest",
    "CreateLog",
    "CreateScoreRequest",
    "CreateSpanRequest",
    "CreateTraceRequest",
    "Error",
    "Event",
    "LlmUsage",
    "Log",
    "MapValue",
    "MethodNotAllowedError",
    "ObservationLevelEvent",
    "ObservationLevelGeneration",
    "ObservationLevelSpan",
    "Score",
    "Span",
    "Trace",
    "TraceIdType",
    "TraceIdTypeEvent",
    "TraceIdTypeGenerations",
    "TraceIdTypeSpan",
    "UnauthorizedError",
    "UpdateGenerationRequest",
    "UpdateSpanRequest",
    "commons",
    "event",
    "generations",
    "score",
    "span",
    "trace",
]



# This file was auto-generated by Fern from our API Definition.

from .create_log import CreateLog
from .llm_usage import LlmUsage
from .log import Log
from .map_value import MapValue
from .observation_level_generation import ObservationLevelGeneration
from .trace_id_type_generations import TraceIdTypeGenerations
from .update_generation_request import UpdateGenerationRequest

__all__ = [
    "CreateLog",
    "LlmUsage",
    "Log",
    "MapValue",
    "ObservationLevelGeneration",
    "TraceIdTypeGenerations",
    "UpdateGenerationRequest",
]



import concurrent.futures
import asyncio
import traceback

from langfuse.log import logger


class FuturesStore:
    def __init__(self):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)
        self.futures = {}
        self.results = {}

    def append(self, id: str, func, *args, future_id=None, **kwargs):
        logger.debug(f"Appending task with ID: {id}")

        if future_id is None:
            logger.debug(f"Task {id} depends on task with ID: {future_id}")
            # If no future_id is provided, store the function and arguments for later execution
            self.futures[id] = (func, args, kwargs)
        else:
            # If a future_id is provided, create a dependent future and store it for later execution
            dependent_future = (func, args, kwargs, future_id)
            self.futures[id] = dependent_future

        return id

    async def flush(self):
        async def run_task(future_id, future_details):
            try:
                logger.debug(f"Running task with ID: {future_id} and details: {future_details}")
                print(f"Running task with ID: {future_id} and details: {future_details}")
                if len(future_details) == 3:  # If there are no dependencies
                    func, args, kwargs = future_details
                    result = await func(*args, **kwargs)
                else:  # If there is a dependency
                    func, args, kwargs, dependent_id = future_details
                    # Wait for the dependency to resolve before running this task
                    if dependent_id not in self.results:
                        await tasks[dependent_id]
                    # get the result from the parent future
                    dependent_result = self.results[dependent_id]
                    result = await func(dependent_result, *args, **kwargs)
                self.results[future_id] = result
            except Exception as e:
                traceback.print_exception(e)
                raise e

        tasks = {}
        final_result = {"status": "success"}
        try:
            # First, create all the tasks but don't run them yet
            for future_id, future_details in self.futures.items():
                if future_id not in self.results:
                    tasks[future_id] = asyncio.create_task(run_task(future_id, future_details))

            # Then, run all the tasks concurrently
            for task in tasks.values():
                await task
        except Exception as e:
            traceback.print_exception(e)
            final_result["status"] = "failed"
            final_result["error"] = str(e)
            raise e

        return final_result



# This file was auto-generated by Fern from our API Definition.

from .errors import AccessDeniedError, Error, MethodNotAllowedError, UnauthorizedError

__all__ = ["AccessDeniedError", "Error", "MethodNotAllowedError", "UnauthorizedError"]



# This file was auto-generated by Fern from our API Definition.

from .types import CreateSpanRequest, ObservationLevelSpan, Span, TraceIdTypeSpan, UpdateSpanRequest

__all__ = ["CreateSpanRequest", "ObservationLevelSpan", "Span", "TraceIdTypeSpan", "UpdateSpanRequest"]



import asyncio
from enum import Enum
import traceback
from typing import Optional
import uuid
from langfuse.api.model import (
    CreateEvent,
    CreateGeneration,
    CreateScore,
    CreateSpan,
    CreateTrace,
    Generation,
    Span,
    UpdateGeneration,
    UpdateSpan,
)
from langfuse.api.resources.event.types.create_event_request import CreateEventRequest
from langfuse.api.resources.generations.types.create_log import CreateLog
from langfuse.api.resources.generations.types.update_generation_request import UpdateGenerationRequest
from langfuse.api.resources.score.types.create_score_request import CreateScoreRequest
from langfuse.api.resources.span.types.create_span_request import CreateSpanRequest
from langfuse.api.resources.span.types.update_span_request import UpdateSpanRequest
from langfuse.futures import FuturesStore
from langfuse.api.client import AsyncFintoLangfuse
from .version import __version__ as version


class Langfuse:
    def __init__(self, public_key: str, secret_key: str, host: Optional[str] = None):
        self.future_store = FuturesStore()

        self.base_url = host if host else "https://cloud.langfuse.com"

        self.client = AsyncFintoLangfuse(
            environment=self.base_url,
            username=public_key,
            password=secret_key,
            x_langfuse_sdk_name="python",
            x_langfuse_sdk_version=version,
        )

        self.future_id = None

    def trace(self, body: CreateTrace):
        try:
            new_id = str(uuid.uuid4())

            async def task(*args):
                try:
                    return await self.client.trace.create(request=body)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.future_store.append(new_id, task, future_id=self.future_id)
            self.future_id = new_id

            return StatefulClient(self.client, None, StateType.TRACE, new_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)

    def span(self, body: Span):
        try:
            new_id = str(uuid.uuid4())

            async def task(*args):
                try:
                    new_body = body.copy(update={"id": new_id})
                    request = CreateSpanRequest(**new_body.dict())
                    return await self.client.span.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.future_store.append(new_id, task, future_id=self.future_id)
            self.future_id = new_id

            return StatefulSpanClient(self.client, new_id, StateType.TRACE, new_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)

    def generation(self, body: Generation):
        try:
            new_id = str(uuid.uuid4()) if body.id is None else body.id
            new_body = body.copy(update={"id": new_id})

            async def task(*args):
                try:
                    request = CreateLog(**new_body.dict())
                    return await self.client.generations.log(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.future_store.append(new_id, task, future_id=self.future_id)
            self.future_id = new_id

            return StatefulGenerationClient(self.client, new_id, StateType.OBSERVATION, new_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)

    async def async_flush(self):
        try:
            return await self.future_store.flush()
        except Exception as e:
            traceback.print_exception(e)

    def flush(self):
        try:
            return asyncio.run(self.future_store.flush())  # Make sure to call self.async_flush() here
        except Exception as e:
            traceback.print_exception(e)


class StateType(Enum):
    OBSERVATION = 1
    TRACE = 0


class StatefulClient:
    def __init__(self, client: Langfuse, id: Optional[str], state_type: StateType, future_id: str, future_store: FuturesStore):
        self.client = client
        self.id = id
        self.future_id = future_id
        self.state_type = state_type
        self.future_store = future_store

    def generation(self, body: CreateGeneration):
        try:
            generation_id = str(uuid.uuid4()) if body.id is None else body.id

            async def task(future_result):
                try:
                    new_body = body.copy(update={"id": generation_id})

                    parent = future_result

                    if self.state_type == StateType.OBSERVATION:
                        new_body = new_body.copy(update={"parent_observation_id": parent.id})
                        new_body = new_body.copy(update={"trace_id": parent.trace_id})
                    else:
                        new_body = new_body.copy(update={"trace_id": parent.id})

                    request = CreateLog(**new_body.dict())
                    return await self.client.generations.log(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            # Add the task to the future store with trace_future_id as a dependency
            self.future_store.append(generation_id, task, future_id=self.future_id)
        except Exception as e:
            traceback.print_exception(e)

        return StatefulGenerationClient(self.client, generation_id, StateType.OBSERVATION, generation_id, future_store=self.future_store)

    def span(self, body: CreateSpan):
        try:
            span_id = str(uuid.uuid4()) if body.id is None else body.id

            async def task(future_result):
                try:
                    new_body = body.copy(update={"id": span_id})

                    parent = future_result
                    print("future_result", future_result)
                    if self.state_type == StateType.OBSERVATION:
                        new_body = new_body.copy(update={"parent_observation_id": parent.id})
                        new_body = new_body.copy(update={"trace_id": parent.trace_id})
                    else:
                        new_body = new_body.copy(update={"trace_id": parent.id})

                    request = CreateSpanRequest(**new_body.dict())
                    return await self.client.span.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            # Add the task to the future store with trace_future_id as a dependency
            self.future_store.append(span_id, task, future_id=self.future_id)

            return StatefulSpanClient(self.client, span_id, StateType.OBSERVATION, span_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)

    def score(self, body: CreateScore):
        try:
            score_id = str(uuid.uuid4()) if body.id is None else body.id

            async def task(future_result):
                try:
                    new_body = body.copy(update={"id": score_id})

                    parent = future_result

                    new_body = body
                    if self.state_type == StateType.OBSERVATION:
                        new_body = new_body.copy(update={"observation_id": parent.id})
                        new_body = new_body.copy(update={"trace_id": parent.trace_id})
                    else:
                        new_body = new_body.copy(update={"trace_id": parent.id})

                    request = CreateScoreRequest(**new_body.dict())
                    return await self.client.score.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.future_store.append(score_id, task, future_id=self.future_id)

            return StatefulClient(self.client, self.id, self.state_type, self.future_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)

    def event(self, body: CreateEvent):
        try:
            event_id = str(uuid.uuid4()) if body.id is None else body.id

            async def task(future_result):
                try:
                    new_body = body.copy(update={"id": event_id})

                    parent = future_result

                    if self.state_type == StateType.OBSERVATION:
                        new_body = new_body.copy(update={"parent_observation_id": parent.id})
                        new_body = new_body.copy(update={"trace_id": parent.trace_id})
                    else:
                        new_body = new_body.copy(update={"trace_id": parent.id})

                    request = CreateEventRequest(**new_body.dict())
                    return await self.client.event.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.future_store.append(body.id, task, future_id=self.future_id)

            return StatefulClient(self.client, event_id, self.state_type, event_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)


class StatefulGenerationClient(StatefulClient):
    def __init__(self, client: Langfuse, id: Optional[str], state_type: StateType, future_id: str, future_store: FuturesStore):
        super().__init__(client, id, state_type, future_id, future_store)

    def update(self, body: UpdateGeneration):
        try:
            future_id = str(uuid.uuid4())
            generation_id = self.future_id

            async def task(future_result):
                try:
                    parent = future_result

                    new_body = body.copy(update={"generation_id": parent.id})

                    request = UpdateGenerationRequest(**new_body.dict())
                    return await self.client.generations.update(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            # Add the task to the future store with trace_future_id as a dependency
            self.future_store.append(future_id, task, future_id=self.future_id)

            return StatefulGenerationClient(self.client, generation_id, StateType.OBSERVATION, future_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)


class StatefulSpanClient(StatefulClient):
    def __init__(self, client: Langfuse, id: Optional[str], state_type: StateType, future_id: str, future_store: FuturesStore):
        super().__init__(client, id, state_type, future_id, future_store)

    def update(self, body: UpdateSpan):
        try:
            future_id = str(uuid.uuid4())
            span_id = self.future_id

            async def task(future_result):
                try:
                    parent = future_result

                    new_body = body.copy(update={"span_id": parent.id})

                    request = UpdateSpanRequest(**new_body.dict())
                    return await self.client.span.update(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.future_store.append(future_id, task, future_id=self.future_id)

            return StatefulSpanClient(self.client, span_id, StateType.OBSERVATION, future_id, future_store=self.future_store)
        except Exception as e:
            traceback.print_exception(e)



# This file was auto-generated by Fern from our API Definition.

from .create_span_request import CreateSpanRequest
from .observation_level_span import ObservationLevelSpan
from .span import Span
from .trace_id_type_span import TraceIdTypeSpan
from .update_span_request import UpdateSpanRequest

__all__ = ["CreateSpanRequest", "ObservationLevelSpan", "Span", "TraceIdTypeSpan", "UpdateSpanRequest"]



# This file was auto-generated by Fern from our API Definition.

from .create_score_request import CreateScoreRequest
from .score import Score
from .trace_id_type import TraceIdType

__all__ = ["CreateScoreRequest", "Score", "TraceIdType"]



# This file was auto-generated by Fern from our API Definition.

from .create_event_request import CreateEventRequest
from .event import Event
from .observation_level_event import ObservationLevelEvent
from .trace_id_type_event import TraceIdTypeEvent

__all__ = ["CreateEventRequest", "Event", "ObservationLevelEvent", "TraceIdTypeEvent"]



import logging
import sys


def setup_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    # Create a console handler and set its level to INFO
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)

    # Create a formatter and set the format for log messages
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(handler)

    return logger


# Use this logger throughout your library
logger = setup_logger()
